/*************************************************************************
*
* Copyright (c) 2012-2023 Barbara Geller & Ansel Sermersheim
* Copyright (c) 2014 Digia Plc
*
* GNU Free Documentation License
* This file may be used under the terms of the GNU Free Documentation
* License version 1.3 as published by the Free Software Foundation
* and appearing in the file included in the packaging of this file.
*
*************************************************************************/

/* classes added here to provide documentation  */

#include <qnamespace.h>
#include <qtconcurrentrun.h>

class QAction;
class QIcon;
class QMetaObject;
class QString;

struct CsLibraryInfo
{
   static constexpr const char *built_for = "";
   static constexpr const char *install_prefix = "";
};


class QCoreApplication
{
   public:
      QCoreApplication ( int & argc, char ** argv );
};

class QColor
{
   public:
      QColor(const QColor &other);
      QColor(QColor &&other);
};

class QMessageBox
{
   public:
      int exec();
};

class QModelIndex
{
   public:
      QModelIndex(const QModelIndex &other);
      QModelIndex(QModelIndex && other);

      ~QModelIndex ();
};

class QObject
{
   public:
      explicit QObject ( QObject * parent = nullptr );
      ~QObject();

      const QMetaObject * metaObject () const;

      static QMetaObject &staticMetaObject();
      static QString tr(const char * text, const char * comment = nullptr, std::optional<int> numArg = std::optional<int>());
};


class QSysInfo
{
   public:
      static QSysInfo::MacVersion macVersion();
};

class QTimer
{
   public:
      template <typename Receiver, typename SlotClass, typename SlotReturn>
      static void singleShot(int msec, Receiver *receiver, SlotReturn (SlotClass::*slotMethod)());

      template <typename Receiver, typename SlotClass, typename SlotReturn>
      static void singleShot(int msec, Qt::TimerType timerType, Receiver *receiver, SlotReturn (SlotClass::*slotMethod)());

      template <typename T>
      static void singleShot(int msec, T slotMethod);

      template <typename T>
      static void singleShot(int msec, Qt::TimerType timerType, T slotMethod);

      template <typename T>
      static void singleShot(int msec, QObject *receiver, T slotMethod);

      template <typename T>
      static void singleShot(int msec, Qt::TimerType timerType, QObject *receiver, T slotMethod);
};

class QToolBar
{
   public:
      template<typename Func1>
      QAction *addAction(const QString &text, const QObject *context, Func1 slot);

      template<typename Func1>
      QAction *addAction(const QIcon &icon, const QString &text, Func1 slot);

      template<typename Func1>
      QAction *addAction(const QIcon &icon, const QString &text, const QObject *context, Func1 slot);
};


// containers

template <typename T>
class QLinkedList<T>::iterator
{
   public:
      using iterator_category = std::bidirectional_iterator_tag;

      iterator();

      iterator ( const iterator & other );
      iterator ( iterator && other );

      iterator & operator= ( const iterator & other );
      iterator & operator= ( iterator && other );

      T & operator* () const;
      T * operator-> () const;

      iterator & operator++ ();
      iterator   operator++ ( int );

      iterator & operator-- ();
      iterator   operator-- ( int );

      bool operator!= ( const_iterator other ) const;
      bool operator== ( const_iterator other ) const;
};

template <typename T>
class QLinkedList<T>::const_iterator
{
   public:
      using iterator_category = std::bidirectional_iterator_tag;

      const_iterator();

      const_iterator ( const const_iterator & other );
      const_iterator ( const iterator & other );

      const_iterator ( const_iterator && other );
      const_iterator ( iterator && other );

      const_iterator & operator= ( const const_iterator & other );
      const_iterator & operator= ( const iterator & other );

      const_iterator & operator= ( const_iterator && other );
      const_iterator & operator= ( iterator && other );


      const T & operator* () const;
      const T * operator-> () const;

      const_iterator & operator++ ();
      const_iterator   operator++ ( int );

      const_iterator & operator-- ();
      const_iterator operator-- ( int );

      bool operator!= ( const_iterator other ) const;
      bool operator== ( const_iterator other ) const;
};

template <typename T>
class QList<T>::iterator
{
   public:
      using iterator_category = std::random_access_iterator_tag;

      iterator();
      iterator ( const iterator & other );
      iterator ( iterator && other );

      iterator & operator= ( const iterator &other );
      iterator & operator= ( iterator && other );

      T &        operator* () const;
      T *        operator-> () const;

      iterator   operator+ ( size_type n ) const;
      iterator & operator++ ();
      iterator   operator++ ( int );
      iterator & operator+= ( size_type n );

      iterator   operator- ( size_type n ) const;
      int        operator- ( iterator other ) const;
      iterator & operator-- ();
      iterator   operator-- ( int );
      iterator & operator-= ( size_type n );

      bool operator< (  const iterator & other ) const;
      bool operator< (  const const_iterator & other ) const;
      bool operator<= ( const iterator & other ) const;
      bool operator<= ( const const_iterator & other ) const;
      bool operator> (  const iterator & other ) const;
      bool operator> (  const const_iterator & other ) const;
      bool operator>= ( const iterator & other ) const;
      bool operator>= ( const const_iterator & other ) const;

      T &  operator[] ( size_type n ) const;

      bool operator!= ( const_iterator other ) const;
      bool operator== ( const_iterator other ) const;
};

template <typename T>
class QList<T>::const_iterator
{
   public:
      using iterator_category = std::random_access_iterator_tag;

      const_iterator();

      const_iterator ( const const_iterator & other );
      const_iterator ( const iterator & other );

      const_iterator ( const_iterator && other );
      const_iterator ( iterator && other );

      const_iterator & operator= ( const const_iterator & other );
      const_iterator & operator= ( const iterator & other );

      const_iterator & operator= ( const_iterator && other );
      const_iterator & operator= ( iterator && other );

      const T &        operator* () const;
      const T *        operator-> () const;

      const_iterator   operator+ ( size_type n ) const;
      const_iterator & operator++ ();
      const_iterator   operator++ ( int );
      const_iterator & operator+= ( size_type n );

      const_iterator   operator- ( size_type n ) const;
      int              operator- ( const_iterator other ) const;
      const_iterator & operator-- ();
      const_iterator   operator-- ( int );
      const_iterator & operator-= ( size_type n );

      bool operator<  ( const iterator & other ) const;
      bool operator<  ( const const_iterator & other ) const;
      bool operator<= ( const iterator & other ) const;
      bool operator<= ( const const_iterator & other ) const;
      bool operator>  ( const iterator & other ) const;
      bool operator>  ( const const_iterator & other ) const;
      bool operator>= ( const iterator & other ) const;
      bool operator>= ( const const_iterator & other ) const;

      const T & operator[] ( size_type n ) const;

      bool operator!= ( const_iterator other ) const;
      bool operator== ( const_iterator other ) const;
};

template <class T1, class T2>
struct QPair {
   using first_type  = T1;
   using second_type = T2;

   QPair()
      : first(), second() {}

   QPair(const T1 &value1, const T2 &value2)
      : first(value1), second(value2)
   {}

   QPair(const QPair & other) = default;
   QPair(QPair && other)      = default;

   QPair<T1, T2> & operator= ( const QPair<T1, T2> & other ) = default;
   QPair<T1, T2> & operator= ( QPair<T1, T2> && other )      = default;

   void swap(QPair &other);

   T1 first;
   T2 second;
};

template <class T1, class T2>
inline bool operator==(const QPair<T1, T2> &p1, const QPair<T1, T2> &p2)
{
   return p1.first == p2.first && p1.second == p2.second;
}

template <class T1, class T2>
inline bool operator!=(const QPair<T1, T2> &p1, const QPair<T1, T2> &p2)
{
   return !(p1 == p2);
}

template <class T1, class T2>
inline bool operator<(const QPair<T1, T2> &p1, const QPair<T1, T2> &p2)
{
   return p1.first < p2.first || (!(p2.first < p1.first) && p1.second < p2.second);
}

template <class T1, class T2>
inline bool operator>(const QPair<T1, T2> &p1, const QPair<T1, T2> &p2)
{
   return p2 < p1;
}

template <class T1, class T2>
inline bool operator<=(const QPair<T1, T2> &p1, const QPair<T1, T2> &p2)
{
   return !(p2 < p1);
}

template <class T1, class T2>
inline bool operator>=(const QPair<T1, T2> &p1, const QPair<T1, T2> &p2)
{
   return !(p1 < p2);
}

class QSet<T>::iterator
{
   public:
      using iterator_category = std::forward_iterator_tag;

      iterator();

      iterator ( const iterator & other );
      iterator ( iterator && other );

      iterator & operator= ( const iterator &other );
      iterator & operator= ( iterator && other );

      const T &  operator* () const;
      iterator   operator+ ( size_type n ) const;
      iterator & operator++ ();
      iterator   operator++ ( int );
      iterator & operator+= ( size_type n );
      const T *  operator-> () const;

      bool operator!= ( iterator other ) const;
      bool operator== ( iterator other ) const;
};

class QSet<T>::const_iterator
{
   public:
      using iterator_category = std::forward_iterator_tag;

      const_iterator();

      const_iterator ( const const_iterator & other );
      const_iterator ( const iterator & other );

      const_iterator ( const_iterator && other );
      const_iterator ( iterator && other );

      const_iterator & operator= ( const const_iterator & other );
      const_iterator & operator= ( const iterator & other );

      const_iterator & operator= ( const_iterator && other );
      const_iterator & operator= ( iterator && other );

      const T &        operator* () const;
      const_iterator   operator+ ( size_type n ) const;
      const_iterator & operator++ ();
      const_iterator   operator++ ( int );
      const_iterator & operator+= ( size_type n );
      const T *        operator-> () const;

      bool operator!= ( const_iterator other ) const;
      bool operator== ( const_iterator other ) const;
};

class QVector<T>::iterator
{
   public:
      using iterator_category = std::random_access_iterator_tag;

      iterator();

      iterator ( const iterator & other );
      iterator ( iterator && other );

      iterator & operator= ( const iterator &other );
      iterator & operator= ( iterator && other );

      T &        operator* () const;
      T *        operator-> () const;

      iterator   operator+ ( size_type n ) const;
      iterator & operator++ ();
      iterator   operator++ ( int );
      iterator & operator+= ( size_type n );

      iterator   operator- ( size_type n ) const;
      size_type  operator- ( iterator other ) const;
      iterator & operator-- ();
      iterator   operator-- ( int );
      iterator & operator-= ( size_type n );

      bool operator< ( const iterator & other ) const;
      bool operator< ( const const_iterator & other ) const;
      bool operator<= ( const iterator & other ) const;
      bool operator<= ( const const_iterator & other ) const;
      bool operator> ( const iterator & other ) const;
      bool operator> ( const const_iterator & other ) const;
      bool operator>= ( const iterator & other ) const;
      bool operator>= ( const const_iterator & other ) const;

      T &  operator[] ( size_type n ) const;

      bool operator!= ( iterator other ) const;
      bool operator== ( iterator other ) const;
};

class QVector<T>::const_iterator
{
   public:
      using iterator_category = std::random_access_iterator_tag;

      const_iterator();

      const_iterator ( const const_iterator & other );
      const_iterator ( const iterator & other );

      const_iterator ( const_iterator && other );
      const_iterator ( iterator && other );

      const_iterator & operator= ( const const_iterator & other );
      const_iterator & operator= ( const iterator & other );

      const_iterator & operator= ( const_iterator && other );
      const_iterator & operator= ( iterator && other );

      const T &        operator* () const;
      const T *        operator-> () const;

      const_iterator   operator+ ( size_type n ) const;
      const_iterator & operator++ ();
      const_iterator   operator++ ( int );
      const_iterator & operator+= ( size_type n );

      const_iterator   operator- ( size_type n ) const;
      size_type        operator- ( const_iterator other ) const;
      const_iterator & operator-- ();
      const_iterator   operator-- ( int );
      const_iterator & operator-= ( size_type n );

      bool operator<  ( const iterator & other ) const;
      bool operator<  ( const const_iterator & other ) const;
      bool operator<= ( const iterator & other ) const;
      bool operator<= ( const const_iterator & other ) const;
      bool operator>  ( const iterator & other ) const;
      bool operator>  ( const const_iterator & other ) const;
      bool operator>= ( const iterator & other ) const;
      bool operator>= ( const const_iterator & other ) const;

      const T & operator[] ( size_type n ) const;

      bool operator!= ( const_iterator other ) const;
      bool operator== ( const_iterator other ) const;
};

namespace QtConcurrent {
   template<class T, class Function, class... Args>
   QFuture<T> run(Function function, Args...);
};


// pointers
template <typename T, typename Deleter = std::default_delete<T>>
class QScopedPointer
{
 public:
   using pointer      = typename std::unique_ptr<T, Deleter>::pointer;
   using element_type = typename std::unique_ptr<T, Deleter>::element_type;
   using deleter_type = typename std::unique_ptr<T, Deleter>::deleter_type;

   using Pointer      = pointer;
   using ElementType  = element_type;
   using DeleterType  = deleter_type;

   explicit QScopedPointer(Pointer p = nullptr) noexcept;
   explicit QScopedPointer(Pointer p, Deleter d) noexcept;
   QScopedPointer(std::unique_ptr<T, Deleter> &&p) noexcept;

   ~QScopedPointer();

   QScopedPointer(QScopedPointer<T, Deleter> && other);
   QScopedPointer &operator=(QScopedPointer<T, Deleter> && other);

   Pointer data () const noexcept;
   Pointer get () const noexcept;
   Deleter & get_deleter() noexcept;
   const Deleter & get_deleter() const noexcept;

   bool isNull () const noexcept;
   Pointer release() noexcept;
   void reset ( Pointer other = nullptr );
   void swap ( QScopedPointer<T, Deleter> & other ) noexcept;
   Pointer take () noexcept;

   explicit operator bool () const noexcept;
   ElementType & operator* () const noexcept;
   Pointer operator-> () const noexcept;
   bool operator! () const noexcept;

   operator std::unique_ptr<T, Deleter>() && noexcept;
};

template <typename T, typename Deleter = std::default_delete<T>>
class QScopedArrayPointer : public QScopedPointer<T, Deleter>
{
 public:
   QScopedArrayPointer( QScopedArrayPointer<element_type> && other ) noexcept;
   QScopedArrayPointer &operator=( QScopedArrayPointer<element_type> && other ) noexcept;

   QScopedArrayPointer( QScopedArrayPointer<element_type[]> && other ) noexcept;
   QScopedArrayPointer &operator=( QScopedArrayPointer<element_type[]> && other ) noexcept;

   T & operator[] ( std::size_t index ) const noexcept;
};


template <typename T, typename Deleter = std::default_delete<T>>
class QUniquePointer
{
};

template <typename T, typename Deleter = std::default_delete<T>>
class QUniqueArrayPointer
{
};


template <typename T>
class QEnableSharedFromThis
{
 public:
   QSharedPointer<T> sharedFromThis();
   QSharedPointer<const T> sharedFromThis() const;
};


template <typename T>
class QSharedPointer
{
 public:
   using element_type = typename std::shared_ptr<T>::element_type;
   using weak_type    = QWeakPointer<T>;
   using pointer      = element_type *;

   using ElementType  = element_type;
   using WeakType     = weak_type;
   using Pointer      = pointer;

   constexpr QSharedPointer() noexcept;
   constexpr QSharedPointer(std::nullptr_t) noexcept;

   template <typename U>
   explicit QSharedPointer(U *p);

   template <typename U, typename Deleter>
   QSharedPointer(U *p, Deleter d);

   template <typename Deleter>
   QSharedPointer(std::nullptr_t, Deleter d);

   template <typename U, typename Deleter, typename Alloc>
   QSharedPointer(U *p, Deleter d, Alloc a);

   template <typename Deleter, typename Alloc>
   QSharedPointer(std::nullptr_t, Deleter d, Alloc a);

   template <typename U, typename Deleter>
   QSharedPointer(QUniquePointer<U, Deleter> &&p);

   template <typename U>
   QSharedPointer(const std::shared_ptr<U> &p) noexcept;

   template <typename U>
   QSharedPointer(std::shared_ptr<U> &&p) noexcept;

   template <typename U>
   QSharedPointer(const QSharedPointer<U> &p1, element_type *p2) noexcept;

   template <typename U>
   QSharedPointer(QSharedPointer<U> &&p1, element_type *p2) noexcept;

   ~QSharedPointer();

   QSharedPointer(const QSharedPointer<T> &other) = default;
   QSharedPointer &operator=(const QSharedPointer<T> &other) = default;

   QSharedPointer(QSharedPointer<T> &&other) = default;
   QSharedPointer &operator=(QSharedPointer<T> && other) = default;

   template <typename U>
   QSharedPointer(const QSharedPointer<U> &p) noexcept;

   template <typename U>
   QSharedPointer &operator=(const QSharedPointer<U> &p) noexcept;

   template <typename U>
   QSharedPointer(QSharedPointer<U> &&p) noexcept;

   template <typename U>
   QSharedPointer &operator=(QSharedPointer<U> &&p) noexcept;

   template <typename U>
   explicit QSharedPointer(const QWeakPointer<U> &p);

   template <typename U>
   QSharedPointer &operator=(const QWeakPointer<U> &p);

   void QSharedPointer::clear ();
   QSharedPointer< T > QSharedPointer::create(Args &&...args);
   T * QSharedPointer::data () const;
   bool QSharedPointer::isNull () const;

   void QSharedPointer::reset();
   void QSharedPointer::reset(U *value);
   void QSharedPointer::reset(U *value, Deleter d);
   void QSharedPointer::reset(U *value, Deleter d, Alloc a);

   void QSharedPointer::swap(QSharedPointer &other);
   QWeakPointer<T> QSharedPointer::toWeakRef () const;

   QSharedPointer::operator bool () const;
   bool QSharedPointer::operator! () const;
   T & QSharedPointer::operator* () const;
   T * QSharedPointer::operator-> () const;

   template <typename U>
   QSharedPointer<U> constCast() const;

   template <typename U>
   QSharedPointer<U> dynamicCast() const;

   template <typename U>
   QSharedPointer<U> objectCast() const;

   template <typename U>
   QSharedPointer<U> staticCast() const;
};

template <typename T>
class QWeakPointer
{
 public:
   using element_type = typename std::shared_ptr<T>::element_type;
   using pointer      = element_type *;

   using ElementType  = element_type;
   using Pointer      = pointer;

   QWeakPointer();

   template <typename U>
   QWeakPointer (const QWeakPointer<U> &other);

   // from QWeakPointer
   QWeakPointer(T *p);

   ~QWeakPointer();

   QWeakPointer(const QWeakPointer<T> &other);
   QWeakPointer &operator=(const QWeakPointer<T> &other);

   QWeakPointer(QWeakPointer<T> && other);
   QWeakPointer &operator=(QWeakPointer<T> && other);

   template <typename U>
   QWeakPointer (const QSharedPointer<U> &p);

   template <typename U>
   QWeakPointer &operator=(const QSharedPointer<U> &p) noexcept;

   bool operator !() const;
   explicit operator bool() const noexcept;

   void clear() noexcept;

   // from QWeakPointer
   Pointer data() const noexcept;

   bool expired() const noexcept;
   bool isNull () const;

   template <typename U>
   bool owner_before(const QSharedPointer<U> &p) const noexcept;

   template <typename U>
   bool owner_before(const QWeakPointer<U> &p) const noexcept;

   void reset() noexcept;
   void swap(QWeakPointer<T> &other) noexcept;
   QSharedPointer<T> toStrongRef () const noexcept;

   long use_count() const noexcept;
};

